import os

import cv2
from mmir.camera import Camera
from mmir.camera.camera import compute_fundamental_matrix
from mmir.feature_detection import GFTTFeatureDetector
from mmir.feature_detection.filters import YPositionFilter, XPositionFilter
from mmir.feature_matching import HogFeatureMatcher
from mmir.mapping import EpipolarTransfer, PolynomialTransformEstimator, MappingVisualizer
from mmir.pipeline import Pipeline
from skimage.io import imread


RSC_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../rsc')


def yield_all_photos():
    left_photos = [os.path.join(RSC_PATH, f) for f in os.listdir(RSC_PATH) if f.endswith("-1.bmp")]
    left_photos.sort()

    right_photos = [os.path.join(RSC_PATH, f) for f in os.listdir(RSC_PATH) if f.endswith("-2.bmp")]
    right_photos.sort()

    flir_photos = [os.path.join(RSC_PATH, f) for f in os.listdir(RSC_PATH) if f.endswith("-7.bmp")]
    flir_photos.sort()

    for (left_photo, right_photo, flir_photo) in zip(left_photos, right_photos, flir_photos):
        yield imread(left_photo), imread(right_photo), imread(flir_photo)


if __name__ == '__main__':
    left_camera = Camera.from_files(
        os.path.join(RSC_PATH, 'left_int.txt'),
        os.path.join(RSC_PATH, 'left_ext.txt'),
    )
    right_camera = Camera.from_files(
        os.path.join(RSC_PATH, 'right_int.txt'),
        os.path.join(RSC_PATH, 'right_ext.txt'),
    )
    flir_camera = Camera.from_files(
        os.path.join(RSC_PATH, 'left_int_F410.txt'),
        os.path.join(RSC_PATH, 'left_ext_F410.txt'),
    )
    left_to_flir = compute_fundamental_matrix(
        left_camera=left_camera,
        right_camera=flir_camera,
    )
    right_to_flir = compute_fundamental_matrix(
        left_camera=right_camera,
        right_camera=flir_camera
    )

    epipolar_transfer = EpipolarTransfer(left_to_flir, right_to_flir)
    pipeline = Pipeline(steps=[
        GFTTFeatureDetector(subregions=16, points_per_region=4, skip_right=True),
        GFTTFeatureDetector.visualize,
        HogFeatureMatcher(
            filters=[
                YPositionFilter(lower_bound=-1, upper_bound=1),
                XPositionFilter(lower_bound=-50, upper_bound=0)
            ],
            best_match_threshold=.2,
        ),
        HogFeatureMatcher.visualize,
        epipolar_transfer,
        PolynomialTransformEstimator(),
        MappingVisualizer(fuse=True)
    ])

    for (left, right, flir) in yield_all_photos():
        epipolar_transfer.destination_image = flir
        pipeline.run(left, right)
        cv2.waitKey(0)



import abc

from dataclasses import dataclass

from mmir.types import Image


@dataclass
class DisparityMap:
    left: Image
    right: Image
    disparity_map: Image


class DisparityMapCalculator(abc.ABC):
    """
    Abstract class to define the behavior of a disparity map computation step.
    """
    @abc.abstractmethod
    def __call__(self, left: Image, right: Image, **kwargs) -> DisparityMap:
        pass








from typing import Callable, Any


class Interceptor:
    """
    Utility class you can use to add in between pipeline steps for various operations like
    swapping data.

    Accepts an intercept function which will be called on the data
    which passes through the interceptor when the pipeline is executed
    """
    def __init__(self, intercept: Callable[[Any], Any]=None):
        self.intercept = intercept

    def __call__(self, *args, **kwargs):
        return self.intercept(*args, **kwargs)



import logging

import cv2
import numpy as np
from skimage.measure import ransac
from skimage.transform import ProjectiveTransform
from skimage.transform._geometric import GeometricTransform

from mmir.feature_matching import Correlations
from mmir.mapping import MappingEstimator, EstimatedMappingModel

LOGGER = logging.getLogger(__name__)


class RansacEstimator(MappingEstimator):
    """
    Uses RANSAC to estimate a given type of model.
    """
    def __init__(
        self,
        model: GeometricTransform = ProjectiveTransform,
        min_samples: int = 4,
        residual_threshold: int = 4,
        visualize: bool=True
    ) -> None:
        self._model = model
        self._min_samples = min_samples
        self._residual_threshold = residual_threshold
        self._visualize = visualize

    def __call__(self, feature_matcher_output: Correlations) -> EstimatedMappingModel:
        left = [l for (l, _, *_) in feature_matcher_output.matches]
        right = [r for (_, r, *_) in feature_matcher_output.matches]

        try:
            model, inliners = ransac(
                (np.array(left), np.array(right)),
                self._model,
                min_samples=self._min_samples,
                residual_threshold=self._residual_threshold,
                max_trials=2000
            )
        except np.linalg.linalg.LinAlgError:
            LOGGER.warning("No convergence...")
            return EstimatedMappingModel(feature_matcher_output.left, feature_matcher_output.right, None)

        filtered_matches = []
        for i, inliner in enumerate(inliners):
            if inliner:
                filtered_matches.append(feature_matcher_output.matches[i])

        if self._visualize:
            left_image = cv2.cvtColor(np.copy(feature_matcher_output.left), cv2.COLOR_GRAY2BGR)
            right_image = cv2.cvtColor(np.copy(feature_matcher_output.right), cv2.COLOR_GRAY2BGR)

            left_image = cv2.resize(src=left_image, dsize=(right_image.shape[1], right_image.shape[0]))

            img = np.concatenate((left_image, right_image), axis=1)

            for i, inliner in enumerate(inliners):
                if not inliner:
                    continue
                pt1 = (int(left[i][0]), int(left[i][1]))
                pt2 = (int(right[i][0] + left_image.shape[1]), int(right[i][1]))

                cv2.line(img=img, pt1=pt1, pt2=pt2, color=(0, 255 if inliner else 0, 0 if inliner else 255))

            cv2.imshow("matches_r", img)

        return EstimatedMappingModel(feature_matcher_output.left, feature_matcher_output.right, model.inverse)



import os
import numpy as np


def transform_point_to_initial_flir(point):
    return point[0] / Sx, point[1] / Sy


IMAGE_SHAPE = (383, 512)

# The scale ratio between the VIS image and the FLIR image.
# Sx = Width_VIS / Width_FLIR
# Sy = Height_VIS / Height_FLIR
Sx = 1.6
Sy = 1.6

# The patch containing the images
BASE_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../rsc')


# The mask we're using for feature detection.
# We're using everything except the dark edges of the FLIR camera, and the FLIR logo.
MASK = np.ones(IMAGE_SHAPE, dtype=np.uint8)
MASK[0:80, 0:512] = 0
MASK[0:383, 470:512] = 0
MASK[0:383, 0:55] = 0
MASK[350:383, 0:512] = 0



from typing import List
import time
from dataclasses import dataclass

from mmir.types import Image, Match
from mmir.math import get_epipolar_line, line_intersection, point_in_bounds
from .disparity_map_calculator import DisparityMap


@dataclass
class OneToOneMapping:
    left: Image
    right: Image
    mapping: List[Match]


class MappingCalculator:
    def __init__(self, left_to_destination, right_to_destination):
        self._left_to_destination = left_to_destination
        self._right_to_destination = right_to_destination

    def __call__(self, feature_matcher_output: DisparityMap) -> OneToOneMapping:
        transfered_disparity_map = []

        for y in range(0, feature_matcher_output.left.shape[0]):
            for x in range(0, feature_matcher_output.left.shape[1]):
                left_point = (x, y)
                right_point = (x - feature_matcher_output.disparity_map[y, x], y)

                epi_line_from_left = get_epipolar_line(self._left_to_destination, left_point)
                epi_line_from_right = get_epipolar_line(self._right_to_destination, right_point)

                point_in_flir = line_intersection(epi_line_from_left, epi_line_from_right)
                if not point_in_bounds(point_in_flir, feature_matcher_output.right.shape):
                    continue
                transfered_disparity_map.append([point_in_flir, (x, y)])

        return OneToOneMapping(
            feature_matcher_output.left,
            feature_matcher_output.right,
            transfered_disparity_map
        )



import numpy as np

from mmir.types import Point, Line, Matrix


def get_epipolar_line(fundamental_matrix: Matrix, point: Point) -> Line:
    """
    Computes the epipolar line, for a given fundamental matrix and point.
    """
    p = np.array([
        point[0],
        point[1],
        1.0
    ], dtype=float)

    return fundamental_matrix @ p



import cv2
import os

from mmir.camera import Camera
from mmir.camera.camera import compute_fundamental_matrix
from mmir.examples.vis_to_flir.config import BASE_PATH, MASK, transform_point_to_initial_flir

from mmir.feature_detection import GFTTFeatureDetector
from mmir.feature_detection.filters import YPositionFilter, XPositionFilter
from mmir.feature_matching import HogFeatureMatcher
from mmir.feature_matching.mi_feature_matcher import MIFeatureMatcher
from mmir.mapping import RansacEstimator, MappingVisualizer
from mmir.pipeline import Pipeline
from skimage.io import imread


def yield_next_images():
    """
    Generator function which returns the next image set to be processed in a form of a tuple
    (vis, flir)
    """
    vis_photos = [os.path.join(BASE_PATH, f) for f in os.listdir(BASE_PATH) if f.endswith("-2.bmp")]
    vis_photos.sort()

    flir_photos = [os.path.join(BASE_PATH, f) for f in os.listdir(BASE_PATH) if f.endswith("-7.bmp")]
    flir_photos.sort()

    for (vis_photo, flir_photo) in zip(vis_photos, flir_photos):
        yield imread(vis_photo), imread(flir_photo)


def preprocess_image_set(vis, flir):
    """
    Preprocessor function for the image set.
    Scales the FLIR photo to the size of the VIS photo, and equalizes the histogram of the VIS
    image.
    """
    flir = cv2.resize(src=flir, dsize=(vis.shape[1], vis.shape[0]))
    return vis, flir


def main():
    flir_camera = Camera.from_files(
        os.path.join(BASE_PATH, 'left_int_F410.txt'),
        os.path.join(BASE_PATH, 'left_ext_F410.txt'),
    )

    vis_camera = Camera.from_files(
        os.path.join(BASE_PATH, 'right_int.txt'),
        os.path.join(BASE_PATH, 'right_ext.txt'),
    )

    fundamental_matrix = compute_fundamental_matrix(
        left_camera=flir_camera,
        right_camera=vis_camera,
    )

    pipline = Pipeline([
        GFTTFeatureDetector(
            subregions=16,
            left_mask=MASK,
            points_per_region=4,
            skip_right=True,
        ),
        GFTTFeatureDetector.visualize,
        MIFeatureMatcher(
            window_size=40,
            filters=[
                YPositionFilter(
                    lower_bound=20,
                    upper_bound=100,
                ),
                XPositionFilter(
                    lower_bound=-32,
                    upper_bound=32,
                ),
            ],
            best_match_threshold=.3,
            matched_point_point_transformer=transform_point_to_initial_flir,
            fundamental_matrix=fundamental_matrix
        ),
        HogFeatureMatcher.visualize,
        RansacEstimator(visualize=True, residual_threshold=4),
        MappingVisualizer(fuse=True),
    ])

    for (right, flir) in yield_next_images():
        (right, flir) = preprocess_image_set(right, flir)
        pipline.run(flir, right)


if __name__ == '__main__':
    main()



from typing import Tuple, List, Union, Optional

import numpy as np

Point = Union[
    Tuple[int, int],
    Tuple[float, float],
    List[int],
    List[float],
    np.array,
]

Line = Union[
    Tuple[int, int, int],
    Tuple[float, float, float],
    List[int],
    List[float],
    np.array,
]

Rectangle = Union[
    Tuple[int, int, int, int],
    Tuple[float, float, float, float],
    List[int],
    List[float],
    np.array,
]

Image = Matrix = Vector = np.array

Match = Tuple[Point, Point, Optional[float]]



from typing import Any

import numpy as np

from mmir.types import Point, Line, Rectangle, Image


def point_to_line_distance(point: Point, line: Line) -> float:
    """
    Computes the distance between a line and a point
    """

    a_r = line[0]
    b_r = line[1]
    c_r = line[2]

    distance = abs(a_r * point[0] + b_r * point[1] + c_r) / ((a_r ** 2 + b_r ** 2) ** 0.5)
    return distance


def line_intersection(line_a: Line, line_b: Line) -> Point:
    """
    Computes the intersection of two lines.
    """
    a = line_a[0]
    b = line_a[1]
    c = line_a[2]

    j = line_b[0]
    k = line_b[1]
    l = line_b[2]

    y = (a * l - c * j) / (b * j - a * k)
    x = (c * k - b * l) / (b * j - a * k)

    return x, y


def ssd(x: np.array, y: np.array) -> float:
    """
    Computes the euclidean distance between two np.arrays.
    """
    return np.linalg.norm(x - y)


def window_in_bounds(center: Point, window_size: int, shape: Rectangle) -> bool:
    """
    Checks if a window centered in a given point, with a given size,
    can fit in a rectangle with a given shape.

    :param center:
        The center of the window
    :param window_size:
        The window's size
    :param shape:
        The shape of the rectangle against which the check is made.
        The expected shape is in the form (rows, cols) as is
        traditional in numpy.
    """

    [x, y] = center

    return not (
        x < window_size / 2 or
        x > shape[1] - window_size / 2 or
        y < window_size / 2 or
        y > shape[0] - window_size / 2
    )


def point_in_bounds(point: Point, shape: Rectangle) -> bool:
    """
    Checks if a point can fit in a rectangle with a given shape.
    """
    [x, y] = point

    return not (
        x < 0 or
        x >= shape[1] or
        y < 0 or
        y >= shape[0]
    )


def roi(center: Point, width: int, height: int, image: Image) -> Image:
    """
    Crops a ROI from an image.
    :param center:
        The center of the ROI.
    :param width:
        The width of the ROI.
    :param height:
        The height of the ROI.
    :param image:
        The image from which to crop the ROI.
    """
    [x, y] = center
    return image[
       int(y - height // 2):int(y + height // 2),
       int(x - width // 2):int(x + width // 2)
    ]


def identity(x: Any) -> Any:
    """
    Identity function. Returns whatever entity it receives.
    """
    return x



import abc
from typing import List, Optional

from dataclasses import dataclass

from mmir.types import Image, Point


@dataclass
class InterestPoints:
    left: Image
    right: Image
    keypoints_left: List[Point]
    keypoints_right: List[Point]


class FeatureDetector(abc.ABC):
    """
    Common interface to be used for the first step of the registration pipeline.

    It defines the input and the output of this pipeline's step.

    It also defines a method for basic visualization of this step.
    """
    def __init__(self, left_mask: Optional[Image]=None, right_mask: Optional[Image]=None):
        """
        :param mask:
            Optional parameter which can be use to mask regions of the image within which
            features should not be computed.
        """
        self._left_mask = left_mask
        self._right_mask = right_mask

    @abc.abstractmethod
    def __call__(self, left_image: Image, right_image: Image) -> InterestPoints:
        pass



import cv2

import numpy as np
from dataclasses import dataclass
from multipledispatch import dispatch

from skimage import transform as tf

from mmir.stereo.disparity_transfer import OneToOneMapping
from .mapping_estimator import EstimatedMappingModel
from mmir.types import Image


@dataclass
class ImageTransformerOutput:
    left: Image
    right: Image
    warped_image: Image


class MappingVisualizer:
    def __init__(self, fuse: bool=False, output_file_path: str=None) -> None:
        self._fuse = fuse
        self._output_file_path = output_file_path

    def _render(self, image: Image):
        if not self._output_file_path:
            cv2.imshow('overlapped' if self._fuse else 'transformed', image)
            cv2.waitKey(0)
            return

        cv2.imwrite(self._output_file_path, image)

    @dispatch(EstimatedMappingModel)
    def __call__(self, mapping_estimator_output: EstimatedMappingModel) -> EstimatedMappingModel:
        warped_image = tf.warp(
            mapping_estimator_output.left,
            mapping_estimator_output.model,
            output_shape=mapping_estimator_output.right.shape

        )

        if self._fuse:
            overlapped = (mapping_estimator_output.right / 255 + warped_image * 3) / 4
            self._render(overlapped)
        else:
            self._render(warped_image)

        return mapping_estimator_output

    @dispatch(OneToOneMapping)
    def __call__(self, disparity_map_calculator_output: OneToOneMapping) -> OneToOneMapping:
        result = np.zeros(disparity_map_calculator_output.left.shape, np.uint8)

        for point in disparity_map_calculator_output.mapping:
            result[(int(point[1][1]), int(point[1][0]))] = \
                disparity_map_calculator_output.right[(int(point[0][1]), int(point[0][0]))]

        if self._fuse:
            overlapped = np.array((result + disparity_map_calculator_output.left / 2) / 1.5, dtype=np.uint8)
            self._render(overlapped)
        else:
            self._render(result)

        return disparity_map_calculator_output




from .epipolar_transfer import EpipolarTransfer
from .mapping_estimator import EstimatedMappingModel, MappingEstimator
from .mapping_visualizer import MappingVisualizer
from .polynomial_transform_estimator import PolynomialTransformEstimator
from .projective_transform_estimator import ProjectiveTransformEstimator
from .ransac_estimator import RansacEstimator



import numpy as np

from skimage import transform as tf

from mmir.feature_matching import Correlations
from mmir.mapping import MappingEstimator, EstimatedMappingModel


class ProjectiveTransformEstimator(MappingEstimator):
    """
    Estimates a polynomial transform from the matches found between a couple of images.
    """
    def __call__(self, feature_matcher_output: Correlations) -> EstimatedMappingModel:
        left = [l for (l, _, _) in feature_matcher_output.matches]
        right = [r for (_, r, _) in feature_matcher_output.matches]

        transformation = tf.ProjectiveTransform()
        transformation.estimate(np.array(right), np.array(left))

        return EstimatedMappingModel(feature_matcher_output.left, feature_matcher_output.right, transformation)
